import numpy as np
from sklearn.model_selection import train_test_split

def _validate_train_val_test_sizes(train_size, val_size, test_size):
    sizes = [train_size, val_size, test_size]
    dtypes = [np.asarray(size).dtype.kind for size in sizes]
    if all(size==None for size in sizes):
        # set default ratios
        train_size = 0.7
        test_size = 0.15
        train_size = 0.15
    elif all(dtype == 'f' for dtype in dtypes):
        sizes_sum = np.sum(sizes)
        if sizes_sum < 0 or sizes_sum > 1:
            raise ValueError(
            'The sum of train_size, test_size and train_size = {}, '
            'should be within (0,1)'.format(sizes_sum))
    elif not all(dtype == 'i' for dtype in dtypes):
        raise ValueError('train_size={}, val_size={}, test_size={} must '
                         'be all integers, all floats or all None'.format(
                             train_size, val_size, test_size))
        
    return tuple(sizes)
    

def train_val_test_split(x:np.ndarray, y:np.ndarray, train_size=None, val_size=None,
                           test_size=None, shuffle=True, random_state=None):
    train_size, val_size, test_size = _validate_train_val_test_sizes(train_size, val_size, test_size)
    x_train, x_val, y_train, y_val = train_test_split(x, y, train_size=train_size, test_size=val_size+test_size,
                                                      shuffle=shuffle, random_state=random_state)
    if np.asarray(val_size).dtype.kind == 'f' and np.asarray(test_size).dtype.kind == 'f':
        normalizer = 1/(val_size+test_size)
        val_size *=normalizer
        test_size *=normalizer
    x_val, x_test, y_val, y_test = train_test_split(x_val, y_val, train_size=val_size,
                                                    test_size=test_size, shuffle=False,
                                                    random_state=random_state)
    return x_train, x_val, x_test, y_train, y_val, y_test

def prepare_train_val_test(x:np.ndarray, y:np.ndarray, train_size=None, val_size=None,
                           test_size=None, preprocessors=None, random_state=None):
    
    x_train, x_val, x_test, y_train, y_val, y_test = \
    train_val_test_split(x, y, train_size, val_size, test_size, random_state)
    
    if preprocessors:
        import sklearn
        for preprocessor in preprocessors:
            if not isinstance(preprocessor, sklearn.base.TransformerMixin):
                raise ValueError('Data preprocessor must be an instance of '
                                 'sklearn.base.TransformerMixin')
            transformer = preprocessor.fit(x_train)
            x_train = transformer.transform(x_train)
            x_val = transformer.transform(x_val)
            x_test = transformer.transform(x_test)
                                           
    return x_train, x_val, x_test, y_train, y_val, y_test  


from quple.components.parameter_table import ParameterTable
from quple.circuit.qubit_register import QubitRegister
from quple.circuit.quantum_circuit import QuantumCircuit
from quple.circuit.parameterised_circuit import ParameterisedCircuit
from quple.circuit.templates.template_circuit_block import TemplateCircuitBlock
from quple.components.descriptors import *
from quple.circuit.common_circuits import *
from quple.utils.utils import *



from typing import List
import copy
import cirq
import quple
import sympy as sp


def replace_symbol_in_op(op, old_symbol:sp.Symbol, new_symbol:sp.Symbol) -> None:
    """Replace the symbol in a parameterised gate with a new symbol."""
    if isinstance(op, cirq.EigenGate):
        if old_symbol in op.exponent.free_symbols:
            op._exponent = op.exponent.subs(old_symbol, new_symbol)

    if isinstance(op, cirq.FSimGate):
        if isinstance(op.theta, sympy.Basic):
            if old_symbol in op.theta.free_symbols:
                op._theta = op.theta.subs(old_symbol, new_symbol)
        if isinstance(op.phi, sympy.Basic):
            if old_symbol in op.phi.free_symbols:
                op._phi = op.phi.subs(old_symbol, new_symbol)

    if isinstance(op, cirq.PhasedXPowGate):
        if isinstance(op.exponent, sympy.Basic):
            if old_symbol in op.exponent.free_symbols:
                op._exponent = op.exponent.subs(old_symbol, new_symbol)
        if isinstance(op.phase_exponent, sympy.Basic):
            if old_symbol in op.phase_exponent.free_symbols:
                op._phase_exponent = op.phase_exponent.subs(old_symbol, new_symbol)
                
def pqc_symbol_map(circuit:cirq.Circuit, symbols_map) -> cirq.Circuit:
    new_circuit = copy.deepcopy(circuit)
    for moment in new_circuit:
        for op in moment:
            if cirq.is_parameterized(op):
                symbols_in_op = quple.symbols_in_op(op.gate)
                for sym in symbols_in_op:
                    replace_symbol_in_op(op.gate, sym, symbols_map[sym])
    return new_circuit



def merge_pqc(circuits:List[cirq.Circuit], symbol:str='θ') -> cirq.Circuit:
    symbol_size = 0
    if not all(isinstance(circuit, cirq.Circuit) for circuit in circuits):
        raise ValueError('Circuits to be merged must be intances of cirq.Circuit object')
    for circuit in circuits:
        symbol_size += len(quple.get_circuit_symbols_in_order(circuit))
    all_symbols = sp.symarray(symbol, symbol_size)
    qubits = set()
    for circuit in circuits:
        qubits |= set(quple.get_circuit_qubits(circuit))
    qubits = sorted(list(qubits))
    merged_circuit = quple.QuantumCircuit(qubits)
    for circuit in circuits:
        old_symbols = quple.get_circuit_symbols_in_order(circuit)
        new_symbols = all_symbols[:len(old_symbols)]
        all_symbols = all_symbols[len(old_symbols):]
        symbols_map = {old:new for old, new in zip(old_symbols, new_symbols)}
        merged_circuit.append(pqc_symbol_map(circuit, symbols_map))
    return merged_circuit
                        
                        


from abc import ABC, abstractmethod
from typing import Optional, List, Union, Callable, Tuple
import numpy as np

import cirq

from quple import ParameterisedCircuit
from quple.circuit.templates.template_circuit_block import TemplateCircuitBlock
from quple.data_encoding.encoding_maps import self_product

class EncodingCircuit(ParameterisedCircuit, ABC):
    '''The data encoding circuit
    A parametrised quantum circuit for encoding classical data into quantum states
    Examples:
    >>> cq = EncodingCircuit(5, depth=1, param_symbol='x')
    '''
    def __init__(self, feature_dimension:int, copies:int=1,
                rotation_blocks:Optional[Union[str, cirq.Gate, Callable, TemplateCircuitBlock,
                                                               List[str],List[cirq.Gate],List[Callable],
                                                               List[TemplateCircuitBlock]]]='H',
                entanglement_blocks:Optional[Union[str, cirq.Gate, Callable, TemplateCircuitBlock,
                                               List[str],List[cirq.Gate],List[Callable],
                                               List[TemplateCircuitBlock]]]=None,
                entangle_strategy:Optional[Union[str,List[str], Callable[[int,int],List[Tuple[int]]],
                                                 List[Callable[[int,int],List[Tuple[int]]]]]]=None,
                encoding_map: Optional[Callable]=None,
                parameter_symbol:str='x',name:str=None,
                *args, **kwargs):
        '''Create a new Pauli expansion circuit.
        Args:
            feature_dimension: dimension of data to be encoded (=number of qubits in the circuit)
            depth: the number of repetition of the encoding circuit
            encoding_map: data mapping function from R^(feature_dimension) to R
            parameter_symbol: the symbol for the unresolved parameter array
            name: name of circuit
        '''
        self._feature_dimension = feature_dimension
        self._encoding_map = encoding_map or self_product
        super().__init__(n_qubit=feature_dimension, copies=copies,
                         rotation_blocks=rotation_blocks,
                         entanglement_blocks=entanglement_blocks,
                         entangle_strategy=entangle_strategy,
                         parameter_symbol=parameter_symbol,
                         final_rotation_layer=False, name=name,
                         *args, **kwargs)
        
    def get_parameters(self, indices:List[int]):
        '''Obtain parameter symbols corresponding to the array indices
        Args:
            indices: array indices of the parameter
        Return:
            parameter symbols
        Example:
        >>>  cq.get_parameters((1,2))
        (x[1], x[2])
        '''
        return self._parameter_table[self._parameter_symbol][indices]

    def encode_parameters(self, indices:List[int]):
        '''Obtain the encoded value for the given parameters using the encoding map
        Args:
            indices: indices of parameter array which corresponds to the indices of qubits in the circuit
        Returns:
            symbolic representation of the encoded data using the encoding map
        Example:
        >>> from quple.data_encoding.PauliExpansion import PauliExpansion
        >>> cq = PauliExpansion(feature_dimension=5, depth=1)
        >>> cq.encode_parameters(indices=(0,2,4))
        >>> (np.pi-x[0])*(np.pi-x[2])*(np.pi-x[4])
        '''
        parameters = self.get_parameters(indices)
        encoded_value = self.encoding_map(parameters)
        return encoded_value
    
    def get_total_parameter_count(self) -> int:
        return self._feature_dimension
    
    @property
    def feature_dimension(self):
        return self._feature_dimension
    
    
    @property
    def encoding_map(self):
        return self._encoding_map
    
    def new_layer(self, name:str=None, *args, **kwargs):
        layer = super().new_layer(name, *args, **kwargs)
        layer._encoding_map = self.encoding_map
        return layer    
    
 


from functools import reduce
import numpy as np
import sympy as sp

from quple.components.interaction_graphs import cyclic

def self_product(x: np.ndarray) -> float:
    """
    Define a function map from R^n to R.

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    coeff = x[0] if len(x) == 1 else reduce(lambda m, n: m * n, x)  
    return coeff
    
def modified_cosine_product(x: np.ndarray) -> float:
    """
    Linear map: f(x) =

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    x[0] = sp.cos(x[0])
    coeff = x[0] if len(x) == 1 else reduce(lambda m, n: m * sp.cos(n), x)  
    return coeff

def cosine_product(x: np.ndarray) -> float:
    """
    Linear map: f(x) =

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    x[0] = sp.cos(x[0]*sp.pi)
    coeff = x[0] if len(x) == 1 else reduce(lambda m, n: m * sp.cos(n), x)  
    return coeff

def distance_measure(x: np.ndarray) -> float:
    """
    Linear map: f(x) =

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    if len(x) == 1:
        coeff = x[0]
    elif len(x) == 2:
        coeff = (x[0] - x[1])/2
    else:
        pairs = cyclic(len(x))
        y = [x[a]-x[b] for a,b in pairs]
        coeff = reduce(lambda m, n: m * n, y)/2**len(pairs)  
    return coeff


def distance_measure(x: np.ndarray) -> float:
    """
    Linear map: f(x) =

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    if len(x) == 1:
        coeff = x[0]
    elif len(x) == 2:
        coeff = (x[0] - x[1])/2
    else:
        pairs = cyclic(len(x))
        y = [x[a]-x[b] for a,b in pairs]
        coeff = reduce(lambda m, n: m * n, y)/2**len(pairs)  
    return coeff

def one_norm_distance(x: np.ndarray) -> float:
    """
    Linear map: f(x) =

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    if len(x) == 1:
        coeff = x[0]
    elif len(x) == 2:
        coeff = sp.abs(x[0] - x[1])
    else:
        pairs = cyclic(len(x))
        y = [sp.abs(x[a]-x[b]) for a,b in pairs]
        coeff = reduce(lambda m, n: m + n, y)/(2*len(pairs))
    return coeff

def two_norm_distance(x: np.ndarray) -> float:
    """
    Linear map: f(x) =

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    if len(x) == 1:
        coeff = x[0]
    elif len(x) == 2:
        coeff = sp.abs(x[0] - x[1])
    else:
        pairs = cyclic(len(x))
        y = [(x[a] - x[b])**2 for a,b in pairs]
        coeff = (reduce(lambda m, n: m + n, y)**(1/2))/np.sqrt(len(pairs))
    return coeff

def arithmetic_mean(x: np.ndarray) -> float:
    """
    Linear map: f(x) =

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    coeff = x[0] if len(x) == 1 else reduce(lambda m, n: m + n, x)/len(x) 
    return coeff    
    



from typing import Optional, Sequence, List, Dict, Union
from abc import ABC
import logging

import numpy as np
import six

import tensorflow as tf
import tensorflow_quantum as tfq

import cirq
import quple
from quple import ParameterisedCircuit

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
# create console handler 
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
logger.addHandler(ch)

from pdb import set_trace


class VQC(tf.keras.Sequential):
    def __init__(self, encoding_circuit:'cirq.Circuit', variational_circuit:'cirq.Circuit', 
                 optimizer:Optional[Union[str,tf.keras.optimizers.Optimizer]]=None, 
                 differentiator:Optional[tfq.differentiators.Differentiator]=None,
                 loss='CategoricalCrossentropy', 
                 activation='sigmoid',
                 metrics=['CategoricalAccuracy'],
                 name:str=None, *arg, **args):
        
        self.encoding_circuit = encoding_circuit
        self.variational_circuit = variational_circuit
        self.differentiator = differentiator        
        self.activation = activation        
        layers = self._get_vqc_layers()
        super(VQC, self).__init__(layers, name)
        optimizer = optimizer or tf.keras.optimizers.Adam()       
        self.compile(loss=loss, optimizer=optimizer, metrics=metrics)

    @property
    def encoding_circuit(self):
        return self._encoding_circuit
    
    @encoding_circuit.setter
    def encoding_circuit(self, val):
        feature_dimension = len(quple.get_circuit_symbols(val))
        if not feature_dimension:
            raise ValueError('Encoding circuit must be a parameterised circuit with '
                             'number of parameters matching the feature dimension')
        self._encoding_circuit = val
        if isinstance(val, quple.QuantumCircuit):
            if val.expr_map is not None:
                feature_dimension = len(quple.symbols_in_expr_map(val.expr_map))
        self._feature_dimension = feature_dimension
        logger.info('Registered encoding circuit with feature dimension: {}'.format(feature_dimension))
        
    @property
    def num_parameters(self) -> int:
        return self.num_parameters
    
    @property
    def feature_dimension(self) -> int:
        return self._feature_dimension
    
    @property
    def variational_circuit(self):
        return self._variational_circuit
    
    @variational_circuit.setter
    def variational_circuit(self, val):
        num_parameters = len(quple.get_circuit_symbols(val))
        if not num_parameters:
            raise ValueError('Variational circuit must be a parameterised circuit which'
                             ' the parameters are to be optimized by the optimizer.')
        self._variational_circuit = val
        self._readout = [cirq.Z(qubit) for qubit in quple.get_circuit_qubits(val)]
        self._num_parameters = num_parameters
        logger.info('Registered variational circuit with number of parameters: {}'.format(num_parameters))
        
    # n_qubit
    
    @property
    def readout(self):
        return self._readout
    
    @property
    def differentiator(self):
        return self._differentiator
    
    @differentiator.setter
    def differentiator(self, val):
        if (val != None) and (not isinstance(val, tfq.differentiators.Differentiator)):
            raise ValueError('Only tensorflow quantum differentiator is allowed')
        self._differentiator = val
    
    @property
    def activation(self):
        return self._activation
    
    @activation.setter
    def activation(self, val):
        if isinstance(val, six.string_types):
            val = tf.keras.activations.get(val)
        self._activation = val
        
    def _get_vqc_layers(self):
        
        input_layer = tf.keras.layers.Input(shape=(), dtype=tf.string)
        pqc_layer = tfq.layers.PQC(self.variational_circuit,
                                   self.readout,
                                   differentiator=self.differentiator)
        output_layer = tf.keras.layers.Dense(1, activation=self.activation)
        
        return [input_layer, pqc_layer, output_layer]
    def _reset_layers(self):
        self._layers = None
        for layer in self._get_vqc_layers():
            self.add(layer)
        
    def _check_data(self, x):
        if isinstance(x, np.ndarray):
            num_dim = x.ndim
            if x.ndim != 2:
                raise ValueError('Data in numpy array format must be of dimension 2')
            num_var = x.shape[1]
            if num_var != self.feature_dimension:
                raise ValueError('Data has feature dimension {} but the encoding'
                ' circuit has feature dimension {}'.format(num_var, self.feature_dimension))
    
    def convert_to_tensor(self, x:np.ndarray):
        self._check_data(x)
        return tfq.convert_to_tensor(self.encoding_circuit.resolve_parameters(x))
        
    def fit(self, x, y,
            batch_size:Optional[int]=None,
            epochs:int=1, 
            verbose:int=1,
            validation_data=None,
            callbacks:Optional[Sequence[tf.keras.callbacks.Callback]]=None,
            *args, **kwargs):
        if isinstance(x, np.ndarray):
            x = self.convert_to_tensor(x)
        if validation_data and isinstance(validation_data, tuple):
            if isinstance(validation_data[0], np.ndarray):
                x_val = self.convert_to_tensor(validation_data[0])
                validation_data = (x_val, validation_data[1])
        
        return super(VQC, self).fit(x, y,
                             batch_size=batch_size,
                             epochs=epochs,
                             validation_data=validation_data,
                             verbose=verbose,
                             callbacks=callbacks, *args, **kwargs)
    
    def evaluate(self, x, y, *args, **kwargs):
        if isinstance(x, np.ndarray):
            x = self.convert_to_tensor(x)
        return super(VQC, self).evaluate(x, y, *args, **kwargs)